#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
"""Generic utility functions for testing.

"""
import os
import sys
import pytest
from contextlib import contextmanager

from atom.api import Atom, Bool

import enaml
from enaml.application import Application, timed_call
from enaml.core.enaml_compiler import EnamlCompiler
from enaml.core.parser import parse
from enaml.widgets.api import Window, Dialog, PopupView
with enaml.imports():
    from enaml.stdlib.message_box import MessageBox

# Timeout for qtbot wait (the value used is large due to Travis being sometimes
# very slow).
TIMEOUT = 2000


def is_qt_available():
    """Check if Qt is installed.

    """
    try:
        import enaml.qt
    except Exception:
        return False
    return True


def compile_source(source, item, filename='<test>', namespace=None):
    """ Compile Enaml source code and return the target item.

    Parameters
    ----------
    source : str
        The Enaml source code string to compile.

    item : str
        The name of the item in the resulting namespace to return.

    filename : str, optional
        The filename to use when compiling the code. The default
        is '<test>'.

    namespace : dict
        Namespace in which to execute the code

    Returns
    -------
    result : object
        The named object from the resulting namespace.

    """
    ast = parse(source, filename)
    code = EnamlCompiler.compile(ast, filename)
    namespace = namespace or {}
    exec(code, namespace)
    return namespace[item]


def run_pending_tasks(qtbot, timeout=1000):
    """Run all enaml pending tasks.

    WARNING: this may not run the Qt event loop if no task is pending.
    This will only deal with tasks scheduled through the schedule function
    (or Application method)

    Parameters
    ----------
    timeout : int, optional
        Timeout after which the operation should fail in ms

    """
    def check_pending_tasks():
        assert not qtbot.enaml_app.has_pending_tasks()
    qtbot.wait_until(check_pending_tasks, TIMEOUT)


def get_window(qtbot, cls=Window, timeout=1000):
    """Convenience function running the event loop and returning the first
    window found in the set of active windows.

    Parameters
    ----------
    cls : type, optional
        Type of the window which should be returned.

    timeout : int, optional
        Timeout after which the operation should fail in ms

    Returns
    -------
    window : Window or None
        Return the first window found matching the specified class

    Raises
    ------
    AssertionError : raised if no window is found in the given time

    """
    def check_window_presence():
        assert [w for w in Window.windows if isinstance(w, cls)]

    qtbot.wait_until(check_window_presence, TIMEOUT)
    for w in Window.windows:
        if isinstance(w, cls):
            return w


def get_popup(qtbot, cls=PopupView, timeout=1000):
    """Convenience function running the event loop and returning the first
    popup found in the set of active popups.

    Parameters
    ----------
    cls : type, optional
        Type of the window which should be returned.

    timeout : int, optional
        Timeout after which the operation should fail in ms

    Returns
    -------
    popup : PopupView or None
        Return the first window found matching the specified class

    Raises
    ------
    AssertionError : raised if no popup is found in the given time

    """
    def check_popup_presence():
        assert [p for p in PopupView.popup_views if isinstance(p, cls)]

    qtbot.wait_until(check_popup_presence, TIMEOUT)
    for p in PopupView.popup_views:
        if isinstance(p, cls):
            return p


def wait_for_window_displayed(qtbot, window, timeout=1000):
    """Wait for a window to be displayed.

    This method should be called on already activated windows (the show method
    should have been called).

    """
    if not window.proxy_is_active or not window.proxy.widget:
        msg = 'Window must be activated before waiting for display'
        raise RuntimeError(msg)
    qtbot.wait_for_window_shown(window.proxy.widget)


class EventObserver(Atom):
    """Simple observer registering the fact it was called once.

    """
    called = Bool()

    def callback(self, change):
        self.called = True

    def assert_called(self):
        assert self.called


def wait_for_destruction(qtbot, widget):
    """Wait for a widget to get destroyed.

    """
    if widget.is_destroyed:
        return
    obs = EventObserver()
    widget.observe('destroyed', obs.callback)
    qtbot.wait_until(obs.assert_called, TIMEOUT)


def close_window_or_popup(qtbot, window_or_popup):
    """Close a window/popup and run the event loop to make sure the closing
    complete.

    """
    if window_or_popup.is_destroyed:
        return
    obs = EventObserver()
    window_or_popup.observe('destroyed', obs.callback)
    window_or_popup.close()
    qtbot.wait_until(obs.assert_called, TIMEOUT)


@contextmanager
def close_all_windows(qtbot):
    """Close all opened windows.

    """
    yield
    run_pending_tasks(qtbot)
    while Window.windows:
        windows = list(Window.windows)
        # First close non top level windows to avoid a window to lose its
        # parent and not remove itself from the set of windows.
        non_top_level_windows = [w for w in windows if w.parent is not None]
        for window in non_top_level_windows:
            close_window_or_popup(qtbot, window)
        for window in windows:
            close_window_or_popup(qtbot, window)


@contextmanager
def close_all_popups(qtbot):
    """Close all opened popups.

    """
    yield
    run_pending_tasks(qtbot)
    while PopupView.popup_views:
        popups = list(PopupView.popup_views)
        # First close non top level popups to avoid a up/window to lose its
        # parent and not remove itself from the set of windows.
        non_top_level_popups = [p for p in popups if p.parent is not None]
        for popup in non_top_level_popups:
            close_window_or_popup(qtbot, popup)
        for popup in popups:
            close_window_or_popup(qtbot, popup)


class ScheduledClosing(object):
    """Scheduled closing of dialog.

    """

    def __init__(self, bot, cls, handler, op, skip_answer):
        self.cls = cls
        self.handler = handler
        self.op = op
        self.bot = bot
        self.skip_answer = skip_answer
        self.called = False

    def __call__(self):
        self.called = True
        from conftest import DIALOG_SLEEP
        dial = get_window(self.bot, cls=self.cls)
        wait_for_window_displayed(self.bot, dial)
        self.bot.wait(DIALOG_SLEEP*1000)
        obs = EventObserver()
        dial.observe('finished', obs.callback)

        try:
            self.handler(self.bot, dial)
        finally:
            if not self.skip_answer:
                getattr(dial, self.op)()
            self.bot.wait_until(obs.assert_called, TIMEOUT)

    def was_called(self):
        """Assert the scheduler was called.

        """
        assert self.called


@contextmanager
def handle_dialog(qtbot, op='accept', handler=lambda qtbot, window: window,
                  cls=Dialog, time=100, skip_answer=False):
    """Automatically close a dialog opened during the context.

    Parameters
    ----------
    op : {'accept', 'reject'}, optional
        Whether to accept or reject the dialog.

    handler : callable, optional
        Callable taking as arguments the bot and the dialog, called before
        accepting or rejecting the dialog.

    cls : type, optional
        Dialog class to identify.

    time : float, optional
        Time to wait before handling the dialog in ms.

    skip_answer : bool, optional
        Skip answering to the dialog. If this is True the handler should handle
        the answer itself.

    """
    sch = ScheduledClosing(qtbot, cls, handler, op, skip_answer)
    timed_call(time, sch)
    try:
        yield
    except Exception:
        raise
    else:
        qtbot.wait_until(sch.was_called, TIMEOUT)


@contextmanager
def handle_question(app, answer):
    """Handle question dialog.

    """
    def answer_question(app, dial):
        """Mark the right button as clicked.

        """
        dial.buttons[0 if answer == 'yes' else 1].was_clicked = True

    with handle_dialog(app, 'accept' if answer == 'yes' else 'reject',
                       handler=answer_question, cls=MessageBox):
        yield


@contextmanager
def cd(path, add_to_sys_path=False):
    """ cd to the directory then return to the cwd

    """
    cwd = os.getcwd()
    if add_to_sys_path:
        abspath = os.path.abspath(path)
        sys.path.append(abspath)
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd)
        if add_to_sys_path:
            sys.path.remove(abspath)


@pytest.fixture
def enaml_run(enaml_qtbot, monkeypatch):
    """ Patches the QtApplication to allow using the qtbot when the
    enaml application is started. It also patches QApplication.exit as
    recommended in the pytest-qt docs.

    Yields
    -------
    handler: object
        an object with a `run` attribute that can be set to a callback that
        will be invoked with the application and first window shown.

    References
    ----------
    1. https://pytest-qt.readthedocs.io/en/latest/app_exit.html

    """
    from enaml.qt.qt_application import QtApplication, QApplication

    app = Application.instance()
    if app:
        Application._instance = None

    class Runner:
        # Set this to a callback
        run = None
    runner = Runner()

    def start(self):
        for window in Window.windows:
            wait_for_window_displayed(enaml_qtbot, window)
            if callable(runner.run):
                runner.run(self, window)
            else:
                close_window_or_popup(enaml_qtbot, window)
            break
    try:
        with monkeypatch.context() as m:
            m.setattr(QtApplication, 'start', start)
            m.setattr(QApplication, 'exit', lambda self: None)
            yield runner
    finally:
        Application._instance = app



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# Copyright (c) 2018, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
import os
import sys
import pytest
import importlib
from utils import is_qt_available, cd, enaml_run


@pytest.mark.skipif(not is_qt_available(), reason='Requires a Qt binding')
def test_workbench(enaml_run, qtbot):
    from enaml.qt.QtCore import Qt

    # Run normally to generate cache files
    dir_path = os.path.abspath(os.path.split(os.path.dirname(__file__))[0])
    example = os.path.join(dir_path, 'examples', 'workbench')

    def handler(app, window):
        widget = window.proxy.widget
        qtbot.wait(1000)
        for i in range(1, 4):
            qtbot.keyClick(widget, str(i), Qt.ControlModifier)
            qtbot.wait(100)
            # TODO: Verify each screen somehow

        qtbot.keyClick(widget, 'q', Qt.ControlModifier)
        # Wait for exit, otherwise it unregisters the commands
        qtbot.wait(100)

    enaml_run.run = handler

    # Add to example folder to the sys path or we get an import error
    with cd(example, add_to_sys_path=True):
        # Now run from cache
        mod = importlib.import_module('sample')
        mod.main()



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from ast import literal_eval
from textwrap import dedent

from utils import compile_source


def test_unicode_literal():
    source = dedent("""\
    from enaml.widgets.api import Window

    enamldef MyWindow(Window): win:
        attr test = u'I do unicode characters: αΨγ 😀⏰🏔 𠁒𠁝𠁖'
    """)
    myWindow = compile_source(source, 'MyWindow')()
    unicode_literal = u'I do unicode characters: αΨγ 😀⏰🏔 𠁒𠁝𠁖'
    assert unicode_literal == myWindow.test


def test_unicode_escape_literal():
    source = dedent("""\
    from enaml.widgets.api import Window

    enamldef MyWindow(Window): win:
        attr test = u'I do unicode characters: \\U00020052'
    """)
    myWindow = compile_source(source, 'MyWindow')()
    unicode_escape_literal = u'I do unicode characters: \U00020052'
    assert unicode_escape_literal == myWindow.test


def test_raw_unicode_literal():
    source = dedent("""\
    from enaml.widgets.api import Window

    enamldef MyWindow(Window): win:
        attr test = r'I do unicode characters: \\U00020052'
    """)
    myWindow = compile_source(source, 'MyWindow')()
    raw_unicode_literal = r'I do unicode characters: \U00020052'
    assert raw_unicode_literal == myWindow.test


def test_bytes_literal():
    source = dedent("""\
    from enaml.widgets.api import Window

    enamldef MyWindow(Window): win:
        attr test = b'I also do bytes with\\nescapes: \\785 and \\x12'
        attr test2 = b'And handle unicode escapes \\u0320 gracefully'
    """)
    myWindow = compile_source(source, 'MyWindow')()
    bytes_literal = b'I also do bytes with\nescapes: \785 and \x12'
    assert bytes_literal == myWindow.test
    bytes_literal = b'And handle unicode escapes \u0320 gracefully'
    assert bytes_literal == myWindow.test2


def test_raw_bytes_literal1():
    source = dedent("""\
    from enaml.widgets.api import Window

    enamldef MyWindow(Window): win:
        attr test = br'I do bytes with\\nescapes'
    """)
    myWindow = compile_source(source, 'MyWindow')()
    raw_bytes_literal = br'I do bytes with\nescapes'
    assert raw_bytes_literal == myWindow.test

def test_raw_bytes_literal2():
    source = dedent("""\
    from enaml.widgets.api import Window

    enamldef MyWindow(Window): win:
        attr test = rb'I do bytes with\\nescapes'
    """)
    myWindow = compile_source(source, 'MyWindow')()
    raw_bytes_literal = literal_eval(r"rb'I do bytes with\nescapes'")
    assert raw_bytes_literal == myWindow.test

def test_unicode_identifier():
    source = dedent("""\
    from enaml.widgets.api import Window

    enamldef MyWindow(Window): win:
        attr 𠁒 = 'test'
        # Check we do not mistake this for an invalid identifier
        attr a = 1.e6
    """)
    myWindow = compile_source(source, 'MyWindow')()
    exec("𠁒 = 'test'")
    assert locals()['𠁒'] == getattr(myWindow, """𠁒""")


if __name__ == '__main__':
    test_unicode_literal()
    test_unicode_escape_literal()
    test_raw_unicode_literal()
    test_bytes_literal()
    test_raw_bytes_literal1()
    test_raw_bytes_literal2()
    test_unicode_identifier()



#------------------------------------------------------------------------------
# Copyright (c) 2019, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
from textwrap import dedent

import pytest

from utils import compile_source, is_qt_available, wait_for_window_displayed


#------------------------------------------------------------------------------
# Attr Syntax
#------------------------------------------------------------------------------
def test_attr_syntax_1():
    source = dedent("""\
    from enaml.widgets.api import *

    enamldef Main(Window):
        attr d
    """)
    compile_source(source, 'Main')


def test_attr_syntax_2():
    source = dedent("""\
    from datetime import date
    from enaml.widgets.api import *

    enamldef Main(Window):
        attr d: date
    """)
    compile_source(source, 'Main')


def test_attr_syntax_3():
    source = dedent("""\
    from datetime import date
    from enaml.widgets.api import *

    enamldef Main(Window):
        attr d: date = date.today()
    """)
    compile_source(source, 'Main')


def test_attr_syntax_4():
    source = dedent("""\
    import datetime
    from enaml.widgets.api import *

    enamldef Main(Window):
        attr d = datetime.date.today()
    """)
    compile_source(source, 'Main')


def test_attr_syntax_5():
    source = dedent("""\
    import datetime
    from enaml.widgets.api import *

    enamldef Main(Window):
        attr d: datetime.date
    """)
    compile_source(source, 'Main')


def test_attr_syntax_6():
    source = dedent("""\
    import datetime
    from enaml.widgets.api import *

    enamldef Main(Window):
        attr d: datetime.date = datetime.date.today()
    """)
    compile_source(source, 'Main')


#------------------------------------------------------------------------------
# Bad attr Syntax
#------------------------------------------------------------------------------
def test_bad_attr_syntax_1():
    # Invalid type
    source = dedent("""\
    import datetime
    from enaml.widgets.api import *

    enamldef Main(Window):
        attr d: datetime.date.now()

    """)
    with pytest.raises(SyntaxError):
        compile_source(source, 'Main')


def test_bad_attr_syntax_2():
    # Dot in attr name
    source = dedent("""\
    from enaml.widgets.api import *

    enamldef Main(Window):
        attr a.d

    """)
    with pytest.raises(SyntaxError):
        compile_source(source, 'Main')


def test_bad_attr_syntax_3():
    # No operator
    source = dedent("""\
    from enaml.widgets.api import *

    enamldef Main(Window):
        attr d: object ""

    """)
    with pytest.raises(SyntaxError):
        compile_source(source, 'Main')

#------------------------------------------------------------------------------
# Typechecks
#------------------------------------------------------------------------------
@pytest.mark.skipif(not is_qt_available(), reason='Requires a Qt binding')
def test_attr_type_1(enaml_qtbot, enaml_sleep):
    import datetime
    source = dedent("""\
    import datetime
    from enaml.widgets.api import *

    enamldef Main(Window):
        attr d: datetime.date = datetime.date.today()
        Container:
            Label:
                text << str(d)
    """)
    tester = compile_source(source, 'Main')()
    tester.show()
    wait_for_window_displayed(enaml_qtbot, tester)

    # Should work
    tester.d = (datetime.datetime.now() - datetime.timedelta(days=1)).date()

    # Should raise a type error
    with pytest.raises(TypeError):
        tester.d = datetime.time(7, 0)



#------------------------------------------------------------------------------
# Copyright (c) 2020, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
"""Test the focus tracker capabilities.

"""
import os

import pytest

from utils import is_qt_available, compile_source, wait_for_window_displayed


SOURCE ="""
from enaml.widgets.api import Window, Field, Container, Feature, FocusTracker


enamldef Main(Window):
    title = 'Focus Tracker'

    alias tracker
    alias f1
    alias f2
    alias f3
    alias f4

    FocusTracker: tracker:
        pass

    Container:
        Field: f1:
            pass
        Field: f2:
            pass
        Field: f3:
            pass
        Field: f4:
            pass

"""

@pytest.mark.skipif("TRAVIS" in os.environ, reason='Skip on Travis')
@pytest.mark.skipif(not is_qt_available(), reason='Requires a Qt binding')
def test_focus_tracking(enaml_qtbot, enaml_sleep):
    """Test moving the focus forward in the presence of a custom focus traversal.

    """
    from enaml.qt import QtCore

    win = compile_source(SOURCE, 'Main')()
    win.show()
    wait_for_window_displayed(enaml_qtbot, win)

    enaml_qtbot.mouseClick(win.f1.proxy.widget, QtCore.Qt.LeftButton)
    assert win.tracker.focused_widget is win.f1

    for w in (win.f2, win.f3, win.f4, win.f1):
        enaml_qtbot.keyClick(win.proxy.widget, QtCore.Qt.Key_Tab)
        assert win.tracker.focused_widget is w

    for w in (win.f4, win.f3, win.f2, win.f1):
        enaml_qtbot.keyClick(
            win.tracker.focused_widget.proxy.widget,
            QtCore.Qt.Key_Tab,
            QtCore.Qt.ShiftModifier
        )
        assert win.tracker.focused_widget is w



#------------------------------------------------------------------------------
# Copyright (c) 2020, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
"""Test winutil icon loading on windows.

"""
import sys
import pytest


@pytest.mark.skipif(sys.platform != "win32", reason="Limited to Windows")
def test_winutil_load_icon():
    from enaml import winutil
    winutil.load_icon(winutil.OIC_INFORMATION)



#------------------------------------------------------------------------------
# Copyright (c) 2018, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#-----------------------------------------------------------------------------
import pytest
from enaml.core.parser import write_tables


@pytest.fixture(scope='session', autouse=True)
def update_parser():
    """Force update of the parser when running the test.

    """
    write_tables()



#------------------------------------------------------------------------------
# Copyright (c) 2018, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
"""Test the handling of nested functions such as lambda functions and
implicit functions used by comprehensions.

"""
import sys

import pytest

from utils import compile_source

SYNCHRONISATION_TEMPLATE =\
"""from enaml.widgets.api import Window, Container, Field, Label

enamldef Main(Window):

    attr colors = ['red', 'blue', 'yellow', 'green']
    alias search_txt : search.text
    alias formatted_comp : lab.text

    Container: container:

        Field: search:
            placeholder = "Search..."
        Label: lab:
            text << f'{colors}'

"""

@pytest.mark.skipif(sys.version_info < (3, 6), reason='Requires Python 3.6')
def test_tracing_fstring():
    """Test that an f-string can be traced.

    """
    source = SYNCHRONISATION_TEMPLATE
    win = compile_source(source, 'Main')()
    assert win.formatted_comp == "['red', 'blue', 'yellow', 'green']"
    win.colors = ['yellow', 'red', 'blue', 'green']
    assert win.formatted_comp == "['yellow', 'red', 'blue', 'green']"


#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
import os
import sys
import types

import pytest

from enaml import imports
from enaml.core.parser import parse
from enaml.core.enaml_compiler import EnamlCompiler
from enaml.widgets.api import PopupView, Window
from utils import (close_window_or_popup, get_popup,
                   wait_for_window_displayed,
                   wait_for_destruction,
                   handle_dialog, handle_question,
                   is_qt_available)

try:
    import numpy
except ImportError:
    NUMPY_AVAILABLE = False
else:
    NUMPY_AVAILABLE = True

try:
    import matplotlib
except ImportError:
    MATPLOTLIB_AVAILABLE = False
else:
    MATPLOTLIB_AVAILABLE = True


try:
    try:
        from PyQt4 import Qsci
    except Exception:
        from PyQt5 import Qsci
except ImportError:
    SCINTILLA_AVAILABLE = False
else:
    SCINTILLA_AVAILABLE = True


try:
    import vtk
except ImportError:
    VTK_AVAILABLE = False
else:
    VTK_AVAILABLE = True


try:
    from enaml.qt import QtWebEngineWidgets
    WEBVIEW_AVAILABLE = True
except Exception:
    WEBVIEW_AVAILABLE = False


if is_qt_available():
    try:
        import enaml.qt.qt_ipython_console
    except ImportError:
        IPY_AVAILABLE = False
    else:
        IPY_AVAILABLE = True
else:
    IPY_AVAILABLE = False


def handle_popup_view_example(qtbot, window):
    """Test showing the popups.

    """
    from conftest import DIALOG_SLEEP

    popup_triggers = window.central_widget().widgets()
    # Test configuration popup
    popup_triggers[0].clicked = True
    popup = get_popup(qtbot)
    qtbot.wait(DIALOG_SLEEP*1000)
    popup.central_widget().widgets()[-1].clicked = True
    wait_for_destruction(qtbot, popup)

    # Test transient popups
    for t in popup_triggers[1:]:
        t.clicked = True
        popup = get_popup(qtbot)
        qtbot.wait(DIALOG_SLEEP*1000)
        close_window_or_popup(qtbot, popup)


def handle_window_closing(qtbot, window):
    """Test answering the question to close the window.

    """
    with handle_question(qtbot, 'yes'):
        window.close()

    def check_window_closed():
        assert not Window.windows

    qtbot.wait_until(check_window_closed)

@pytest.mark.skipif(not is_qt_available(), reason='Requires a Qt binding')
@pytest.mark.parametrize("path, handler",
                         [('aliases/chained_attribute_alias.enaml', None),
                          ('aliases/chained_widget_alias.enaml', None),
                          ('aliases/simple_attribute_alias.enaml', None),
                          ('aliases/simple_widget_alias.enaml', None),
                          pytest.param('applib/live_editor.enaml', None,
                                       marks=pytest.mark.skipif(
                                           not SCINTILLA_AVAILABLE,
                                           reason='Requires scintilla')),
                          ('dynamic/conditional.enaml', None),
                          ('dynamic/fields.enaml', None),
                          ('dynamic/looper.enaml', None),
                          ('dynamic/notebook_pages.enaml', None),
                          ('functions/declare_function.enaml', None),
                          ('functions/observe_model_signal.enaml', None),
                          ('functions/override_function.enaml', None),
                          ('layout/basic/align_offset.enaml', None),
                          ('layout/basic/align.enaml', None),
                          ('layout/basic/grid.enaml', None),
                          ('layout/basic/hbox_equal_widths.enaml', None),
                          ('layout/basic/hbox_spacing.enaml', None),
                          ('layout/basic/hbox.enaml', None),
                          ('layout/basic/horizontal.enaml', None),
                          ('layout/basic/linear_relations.enaml', None),
                          ('layout/basic/vbox.enaml', None),
                          ('layout/basic/vertical.enaml', None),
                          pytest.param('layout/advanced/button_ring.enaml',
                                       None,
                                       marks=pytest.mark.skipif(
                                           not (NUMPY_AVAILABLE and
                                                is_qt_available()),
                                           reason='Requires numpy and Qt'),),
                          ('layout/advanced/factory_func.enaml', None),
                          ('layout/advanced/find_replace.enaml', None),
                          ('layout/advanced/fluid.enaml', None),
                          ('layout/advanced/manual_hbox.enaml', None),
                          ('layout/advanced/manual_vbox.enaml', None),
                          ('layout/advanced/nested_boxes.enaml', None),
                          ('layout/advanced/nested_containers.enaml', None),
                          ('layout/advanced/override_layout_constraints.enaml',
                           None),
                          ('stdlib/mapped_view.enaml', None),
                          ('stdlib/message_box.enaml', None),
                          ('stdlib/slider_transform.enaml', None),
                          ('stdlib/task_dialog.enaml', None),
                          ('styling/banner.enaml', None),
                          ('styling/dock_item_alerts.enaml', None),
                          ('styling/gradient_push_button.enaml', None),
                          ('templates/basic.enaml', None),
                          ('templates/advanced.enaml', None),
                          ('widgets/buttons.enaml', None),
                          ('widgets/context_menu.enaml', None),
                          ('widgets/dock_area.enaml', None),
                          ('widgets/dock_pane.enaml', None),
                          ('widgets/dual_slider.enaml', None),
                          ('widgets/file_dialog.enaml', None),
                          ('widgets/flow_area.enaml', None),
                          ('widgets/focus_traversal.enaml', None),
                          ('widgets/form_spacing.enaml', None),
                          ('widgets/form.enaml', None),
                          ('widgets/group_box.enaml', None),
                          ('widgets/h_group.enaml', None),
                          pytest.param('widgets/ipython_console.enaml', None,
                                       marks=pytest.mark.skipif(
                                           not IPY_AVAILABLE,
                                           reason=('Requires IPython and its '
                                                   'qt console.'))),
                          ('widgets/image_view.enaml', None),
                          ('widgets/main_window.enaml', None),
                          ('widgets/menu_bar.enaml', None),
                          pytest.param('widgets/mpl_canvas.enaml', None,
                                       marks=pytest.mark.skipif(
                                           not MATPLOTLIB_AVAILABLE,
                                           reason='Requires matplotlib')),
                          ('widgets/notebook.enaml', None),
                          ('widgets/popup_menu.enaml', None),
                          ('widgets/popup_view.enaml',
                           handle_popup_view_example),
                          ('widgets/progress_bar.enaml', None),
                          ('widgets/scroll_area.enaml', None),
                          ('widgets/slider.enaml', None),
                          ('widgets/spin_box.enaml', None),
                          ('widgets/splitter.enaml', None),
                          ('widgets/tool_bar.enaml', None),
                          ('widgets/tool_buttons.enaml', None),
                          ('widgets/v_group.enaml', None),
                          pytest.param('widgets/vtk_canvas.enaml', None,
                                       marks=pytest.mark.skipif(
                                           not VTK_AVAILABLE,
                                           reason='Requires vtk')),
                          pytest.param('widgets/web_view.enaml', None,
                                       marks=pytest.mark.skipif(
                                           not WEBVIEW_AVAILABLE,
                                           reason='Requires webview')),
                          ('widgets/window_children.enaml', None),
                          ('widgets/window_closing.enaml',
                           handle_window_closing),
                          ('widgets/window.enaml', None)])
def test_examples(enaml_qtbot, enaml_sleep, path, handler):
    """ Test the enaml examples.

    """
    dir_path = os.path.abspath(os.path.split(os.path.dirname(__file__))[0])
    enaml_file = os.path.join(dir_path, 'examples', os.path.normpath(path))

    with open(enaml_file, 'r') as f:
        enaml_code = f.read()

    # Parse and compile the Enaml source into a code object
    ast = parse(enaml_code, filename=enaml_file)
    code = EnamlCompiler.compile(ast, enaml_file)

    # Create a proper module in which to execute the compiled code so
    # that exceptions get reported with better meaning
    try:
        module = types.ModuleType('enaml_test')
        module.__file__ = os.path.abspath(enaml_file)
        sys.modules['enaml_test'] = module
        ns = module.__dict__

        # Put the directory of the Enaml file first in the path so relative
        # imports can work.
        sys.path.insert(0, os.path.abspath(os.path.dirname(enaml_file)))
        with imports():
            exec(code, ns)

        window = ns['Main']()
        window.show()
        window.send_to_front()
        wait_for_window_displayed(enaml_qtbot, window)
        enaml_qtbot.wait(enaml_sleep*1000)

        if handler is not None:
            handler(enaml_qtbot, window)

    finally:
        # Make sure we clean up the sys modification before leaving
        sys.path.pop(0)
        del sys.modules['enaml_test']



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# Copyright (c) 2013-2018, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
import os
import sys
import shutil
import pytest
import importlib
from enaml.compile_all import compileall
from utils import cd, enaml_run


def clean_cache(path):
    """ Clean the cache files in the path

    """
    with cd(path):
        if os.path.exists('__enamlcache__'):
            shutil.rmtree('__enamlcache__')
        if os.path.exists('__pycache__'):
            shutil.rmtree('__pycache__')
        for f in os.listdir('.'):
            if f.endswith('.pyc'):
                os.remove(f)


def clean_source(path):
    """ Clean the source files in the path

    """
    with cd(path):
        for f in os.listdir('.'):
            if os.path.splitext(f)[-1] in (
                        '.py',
                        '.enaml',
                    ):
                os.remove(f)


@pytest.mark.parametrize("tutorial", [
    'employee',
    'hello_world',
    'person'
])
def test_tutorials(enaml_run, tmpdir, tutorial):
    # Run normally to generate cache files

    dir_path = os.path.abspath(os.path.split(os.path.dirname(__file__))[0])
    source = os.path.join(dir_path, 'examples', 'tutorial', tutorial)
    example = os.path.join(tmpdir.strpath, tutorial)

    # Copy to a tmp dir
    shutil.copytree(source, example)
    clean_cache(example)  # To be safe

    # Run compileall
    compileall.compile_dir(example)

    # Remove source files
    clean_source(example)

    # Add to example folder to the sys path or we get an import error
    with cd(example, add_to_sys_path=True):
        # Python only uses pyc files if copied from the pycache folder
        for f in os.listdir('__pycache__'):
            cf = ".".join(f.split(".")[:-2]) + ".pyc"
            shutil.copy(os.path.join('__pycache__', f), cf)

        # Verify it's clean
        assert not os.path.exists(tutorial+".py")
        assert not os.path.exists(tutorial+"_view.enaml")

        # Now run from cache
        mod = importlib.import_module(tutorial)
        mod.main()



#------------------------------------------------------------------------------
# Copyright (c) 2013-2018, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
import gc
from textwrap import dedent
from collections import Counter

import pytest

from enaml.compat import PY39
from enaml.core.declarative_function import (DeclarativeFunction,
                                             BoundDeclarativeMethod)
from utils import compile_source


def test_declarative_function():
    """Test the descriptors and repr of DeclarativeFunction.

    """
    def test(a, b=1):
        pass

    dfunc = DeclarativeFunction(test, '_scope_key')
    assert dfunc.__func__ is test
    assert dfunc.__key__ == '_scope_key'
    assert dfunc._d_func
    assert 'declarative function' in repr(dfunc)

    with pytest.raises(TypeError) as excinfo:
        DeclarativeFunction(1, 1)
    assert 'function' in excinfo.exconly()


def test_bound_declarative_method():
    """Test the descriptors and repr of BoundDeclarativeMethod.

    """
    def test(a, b=1):
        pass

    SELF = object()

    bdmethod = DeclarativeFunction(test, '_scope_key').__get__(SELF, None)
    assert bdmethod.__func__ is test
    assert bdmethod.__key__ == '_scope_key'
    assert bdmethod.__self__ is SELF
    assert 'bound declarative method' in repr(bdmethod)


def test_declarative_function_get_and_call():
    """Test calling DeclarativeFunction and BoundDeclarativeMethod.

    """
    source = dedent("""\
    from enaml.widgets.window import Window

    enamldef MyWindow(Window): main:

        func call(arg, kwarg=1):
            assert main is self
            return main

        func call2():
            return main

    """)
    tester = compile_source(source, 'MyWindow')()
    bdmethod = tester.call
    assert isinstance(bdmethod, BoundDeclarativeMethod)
    for i in range(100):
        assert bdmethod(1) is tester
        assert bdmethod(arg=2) is tester
        assert bdmethod(1, kwarg=1) is tester
    func = type(tester).call
    assert isinstance(func, DeclarativeFunction)
    for i in range(100):
        assert func(tester, 1) is tester
        assert func(tester, arg=1) is tester
        assert func(tester, arg=1, kwarg=2) is tester
    func = type(tester).call2
    assert isinstance(func, DeclarativeFunction)
    for i in range(100):
        assert func(tester) is tester
        assert func(tester) is tester

    with pytest.raises(TypeError) as excinfo:
        type(tester).call2()
    assert 'DeclarativeFunction.__call__()' in excinfo.exconly()
    assert '1 argument' in excinfo.exconly()


def test_calling_super():
    """Test calling a declarative function using super.

    """
    source = dedent("""\
    from enaml.widgets.window import Window

    enamldef MyWindow(Window): main:

        func call(arg, kwarg=1):
            return super()

    """)
    tester = compile_source(source, 'MyWindow')()
    with pytest.raises(TypeError) as excinfo:
        tester.call(1)
    assert 'super()' in excinfo.exconly()


@pytest.mark.parametrize('attribute',
                         ['__builtins__', '_d_storage'])
def test_invoking_broken_dfunc(attribute):
    """Test invoking a declarative function whose one key attribute is missing.

    """
    source = dedent("""\
    from enaml.widgets.window import Window

    enamldef MyWindow(Window): main:

        func call(arg, kwarg=1):
            return 1

    """)
    tester = compile_source(source, 'MyWindow')()
    if attribute == '__builtins__':
        del tester.call.__func__.__globals__['__builtins__']
        obj = tester
    elif attribute == '_d_storage':
        del tester.call.__self__._d_storage
        obj = object()
    else:
        raise RuntimeError('wrong parameter')
    exc = KeyError if attribute == '__builtins__' else AttributeError
    with pytest.raises(exc) as excinfo:
        type(tester).call(obj, 1)
    assert attribute in excinfo.exconly()


def test_traversing_bound_method():
    """Test traversing a bound method.

    """
    source = dedent("""\
    from enaml.widgets.window import Window

    enamldef MyWindow(Window): main:

        func call(arg, kwarg=1):
            return super()

    """)
    tester = compile_source(source, 'MyWindow')()
    assert (Counter(gc.get_referents(tester.call)) ==
        Counter([tester.call.__func__, tester, tester.call.__key__] +
                ([type(tester.call)] if PY39 else []))
    )


if __name__ == '__main__':
    test_bound_declarative_method()
    test_declarative_function()
    test_declarative_function_get_and_call()



#------------------------------------------------------------------------------
# Copyright (c) 2020, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#------------------------------------------------------------------------------
import traceback as tb
from textwrap import dedent

import pytest

from utils import compile_source


def test_validate_declarative_1():
    """ Test that we reject children that are not type in enamldef.

    This also serves to test the good working of try_squash_raise.

    """
    source = dedent("""\
    from enaml.widgets.api import *

    a = 1

    enamldef Main(Window):
        a:
            pass

    """)
    with pytest.raises(TypeError) as exc:
        Main = compile_source(source, 'Main')

    ftb = "\n".join(tb.format_tb(exc.tb))
    assert " validate_declarative" not in ftb


def test_validate_declarative_2():
    """ Test that we reject children that are not declarative in enamldef.

    This also serves to test the good working of try_squash_raise.

    """
    source = dedent("""\
    from enaml.widgets.api import *

    class A:
        pass

    enamldef Main(Window):
        A:
            pass

    """)
    with pytest.raises(TypeError) as exc:
        Main = compile_source(source, 'Main')

    ftb = "\n".join(tb.format_tb(exc.tb))
    assert " validate_declarative" not in ftb



#------------------------------------------------------------------------------
# Copyright (c) 2013-2020, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
import os
import sys
import inspect
from setuptools import find_packages, Extension, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install
from setuptools.command.develop import develop

sys.path.insert(0, os.path.abspath('.'))
from enaml.version import __version__

# Use the env var ENAML_DISABLE_FH4 to disable linking against VCRUNTIME140_1.dll

ext_modules = [
    Extension(
        'enaml.weakmethod',
        ['enaml/src/weakmethod.cpp'],
        language='c++',
    ),
    Extension(
        'enaml.callableref',
        ['enaml/src/callableref.cpp'],
        language='c++',
    ),
    Extension(
       'enaml.signaling',
       ['enaml/src/signaling.cpp'],
       language='c++',
    ),
    Extension(
        'enaml.core.funchelper',
        ['enaml/src/funchelper.cpp'],
        language='c++',
    ),
    Extension(
        'enaml.colorext',
        ['enaml/src/colorext.cpp'],
        language='c++',
    ),
    Extension(
        'enaml.fontext',
        ['enaml/src/fontext.cpp'],
        language='c++',
    ),
    Extension(
        'enaml.core.dynamicscope',
        ['enaml/src/dynamicscope.cpp'],
        language='c++',
    ),
    Extension(
        'enaml.core.alias',
        ['enaml/src/alias.cpp'],
        language='c++',
    ),
    Extension(
        'enaml.core.declarative_function',
        ['enaml/src/declarative_function.cpp'],
        language='c++',
    ),
]


if sys.platform == 'win32':
    ext_modules.append(
        Extension(
            'enaml.winutil',
            ['enaml/src/winutil.cpp'],
            libraries=['user32', 'gdi32'],
            language='c++'
        )
    )


class BuildExt(build_ext):
    """ A custom build extension for adding compiler-specific options.

    """
    c_opts = {
        'msvc': ['/EHsc']
    }

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.debug = False

    def build_extensions(self):

        # Delayed import of cppy to let setup_requires install it if necessary
        import cppy

        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        for ext in self.extensions:
            ext.include_dirs.insert(0, cppy.get_include())
            ext.extra_compile_args = opts
            if sys.platform == 'darwin':
                ext.extra_compile_args += ['-stdlib=libc++']
                ext.extra_link_args += ['-stdlib=libc++']
            if (ct == 'msvc' and os.environ.get('ENAML_DISABLE_FH4')):
                # Disable FH4 Exception Handling implementation so that we don't
                # require VCRUNTIME140_1.dll. For more details, see:
                # https://devblogs.microsoft.com/cppblog/making-cpp-exception-handling-smaller-x64/
                # https://github.com/joerick/cibuildwheel/issues/423#issuecomment-677763904
                ext.extra_compile_args.append('/d2FH4-')
        build_ext.build_extensions(self)


class Install(install):
    """ Calls the parser to construct a lex and parse table specific
        to the system before installation.

    """

    def run(self):
        try:
            from enaml.core.parser import write_tables
            write_tables()
        except ImportError:
            pass
        # Follow logic used in setuptools
        # cf https://github.com/pypa/setuptools/blob/master/setuptools/command/install.py#L58
        if self.old_and_unmanageable or self.single_version_externally_managed:
            return install.run(self)

        if not self._called_from_setup(inspect.currentframe()):
            # Run in backward-compatibility mode to support bdist_* commands.
            install.run(self)
        else:
            self.do_egg_install()


class Develop(develop):
    """ Calls the parser to construct a lex and parse table specific
        to the system before installation.

    """

    def run(self):
        try:
            from enaml.core.parser import write_tables
            write_tables()
        except ImportError:
            pass
        develop.run(self)


setup(
    name='enaml',
    version=__version__,
    author='The Nucleic Development Team',
    author_email='sccolbert@gmail.com',
    url='https://github.com/nucleic/enaml',
    description='Declarative DSL for building rich user interfaces in Python',
    long_description=open('README.rst').read(),
    license='BSD',
    classifiers=[
          # https://pypi.org/pypi?%3Aaction=list_classifiers
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: Implementation :: CPython',
      ],
    python_requires='>=3.6',
    requires=['atom', 'qtpy', 'ply', 'kiwisolver'],
    install_requires=['atom>=0.6.0',
                      'kiwisolver>=1.2.0',
                      'ply>=3.4',
                      "bytecode>=0.11.0"
                      ],
    setup_requires=['cppy>=1.1.0'],
    extras_require={
        "qt5-pyqt": ["qtpy", "pyqt5"],
        "qt5-pyside": ["qtpy", "pyside2"]
    },
    packages=find_packages(),
    package_data={
        'enaml.applib': ['*.enaml'],
        'enaml.stdlib': ['*.enaml'],
        'enaml.workbench.core': ['*.enaml'],
        'enaml.workbench.ui': ['*.enaml'],
        'enaml.qt.docking': [
            'dock_images/*.png',
            'dock_images/*.py',
            'enaml_dock_resources.qrc'
        ],
    },
    entry_points={'console_scripts': [
        'enaml-run = enaml.runner:main',
        'enaml-compileall = enaml.compile_all:main',
    ]},
    ext_modules=ext_modules,
    cmdclass={'build_ext': BuildExt,
              'install': Install,
              'develop': Develop},
)



#------------------------------------------------------------------------------
# Copyright (c) 2013-2017, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
#------------------------------------------------------------------------------
import pytest
from utils import compile_source, wait_for_window_displayed,is_qt_available

pytestmark = pytest.mark.skipif(not is_qt_available(),
                                reason='Requires a Qt binding')

MANUAL_REPARENTING = \
"""from enaml.core.api import Conditional
from enaml.widgets.api import (Window, Container, ObjectCombo, PushButton,
                               ScrollArea, GroupBox, Label)


enamldef Displayable(GroupBox): cont:

    attr sp_children = []
    alias btn_clicked : btn.clicked

    func refresh():
        for c in sp_children:
            c.set_parent(self)
            c.show()

    initialized ::
        refresh()

    PushButton: btn:
        text = 'Add label'
        clicked ::
            Label(text='Dummy').set_parent(cont)

C2 = Displayable(title='C2')

C1 = Displayable(title='C1', sp_children=[C2])


enamldef Main(Window):

    attr displayables = {'C1': C1, 'C2': C2}
    alias selected_disp : cb.selected

    initialized::
        C1.set_parent(scroll)
        C1.show()

    Container:
        ObjectCombo: cb:
            items = sorted(displayables)
            selected ::
                view = displayables[change['value']]
                view.set_parent(scroll)
                view.refresh()
                view.show()

        ScrollArea: scroll:
            pass
"""


def test_manual_reparenting(enaml_qtbot):
    """Test manually reparenting widgets which are previously child and parent.

    """
    win = compile_source(MANUAL_REPARENTING, 'Main')()
    win.show()
    win.send_to_front()
    wait_for_window_displayed(enaml_qtbot, win)
    win.selected_disp = 'C2'
    win.displayables['C2'].btn_clicked = True
